/*
  ==============================================================================

    This file was auto-generated!

    It contains the basic startup code for a Juce application.

  ==============================================================================
*/

#include "PluginProcessor.h"
#include "PluginEditor.h"


//==============================================================================
FnordAudioProcessor::FnordAudioProcessor()
{
	Names.add("Interp");
	Names.add("Size");
	Names.add("Algo");
	Names.add("Change");
	Names.add("Attack");
	Names.add("Sustain");
	Names.add("Release");
	Names.add("Poly");
	Names.add("Vol");
	setParameterNotifyingHost(Interp, 0.0f);
	setParameterNotifyingHost(Size, 1.0f);
	setParameterNotifyingHost(Algo, 0.0f);
	setParameterNotifyingHost(Change, .75f);
	setParameterNotifyingHost(Attack, 0.0f);
	setParameterNotifyingHost(Sustain, 1.0f);
	setParameterNotifyingHost(Release, 0.1f);
	setParameterNotifyingHost(Poly, 1.0f);
	setParameterNotifyingHost(Vol, .75f);
	for (int i = 0; i < VOICES; ++i)
	{
		mNote[i] = 0;
		mVelo[i] = 0;
		mOrder[i] = 0;
		mEnvStage[i] = Off;
		mEnv[i] = 0;
		mStep[i] = 0;
		mPhase[i] = 0;
	}
	mNoteCount = 0;

	for (int i = 0; i < MAX_TABLE; ++i)
	{
		table[i] = (Random::getSystemRandom().nextFloat()*2.0f)-1.0f;
	}
}

FnordAudioProcessor::~FnordAudioProcessor()
{
}

//==============================================================================
const String FnordAudioProcessor::getName() const
{
    return JucePlugin_Name;
}

int FnordAudioProcessor::getNumParameters()
{
	return numParams;
}

float FnordAudioProcessor::getParameter (int index)
{
	if (index >= 0 && index < numParams)
	{
		return Params[index];
	}
	return 0.0f;
}

void FnordAudioProcessor::setParameter (int index, float newValue)
{
	if (index >= 0 && index < numParams)
	{
		Params[index] = newValue;
	}
}

const String FnordAudioProcessor::getParameterName (int index)
{
	if (index < numParams)
	{
		return Names[index];
	}
	return String::empty;
}

const String FnordAudioProcessor::getParameterText (int index)
{
	if (index == Size)
	{
		return String(getSize(Params[Size]));
	}
	if (index == Poly) {
		return String(getPoly());
	}
	if (index == Algo) {
		return getAlgoName();
	}
	return String(getParameter(index));
}

const String FnordAudioProcessor::getInputChannelName (int channelIndex) const
{
	return String(channelIndex + 1);
}

const String FnordAudioProcessor::getOutputChannelName (int channelIndex) const
{
	return String(channelIndex + 1);
}

bool FnordAudioProcessor::isInputChannelStereoPair (int /*index*/) const
{
    return true;
}

bool FnordAudioProcessor::isOutputChannelStereoPair (int /*index*/) const
{
    return true;
}

bool FnordAudioProcessor::acceptsMidi() const
{
   #if JucePlugin_WantsMidiInput
    return true;
   #else
    return false;
   #endif
}

bool FnordAudioProcessor::producesMidi() const
{
   #if JucePlugin_ProducesMidiOutput
    return true;
   #else
    return false;
   #endif
}

bool FnordAudioProcessor::silenceInProducesSilenceOut() const
{
    return false;
}

double FnordAudioProcessor::getTailLengthSeconds() const
{
    return 0.0;
}

int FnordAudioProcessor::getNumPrograms()
{
    return 1;
}

int FnordAudioProcessor::getCurrentProgram()
{
    return 0;
}

void FnordAudioProcessor::setCurrentProgram (int /*index*/)
{
}

const String FnordAudioProcessor::getProgramName (int /*index*/)
{
    return String::empty;
}

void FnordAudioProcessor::changeProgramName (int /*index*/, const String& /*newName*/)
{
}

//==============================================================================
void FnordAudioProcessor::prepareToPlay (double sampleRate, int /*samplesPerBlock*/)
{
	mSR = sampleRate;
}

void FnordAudioProcessor::releaseResources()
{
    // When playback stops, you can use this as an opportunity to free up any
    // spare memory, etc.
}

//==============================================================================
bool FnordAudioProcessor::hasEditor() const
{
    return false; // (change this to false if you choose to not supply an editor)
}

AudioProcessorEditor* FnordAudioProcessor::createEditor()
{
    return new FnordAudioProcessorEditor (this);
}

//==============================================================================
void FnordAudioProcessor::getStateInformation (MemoryBlock& destData)
{
	XmlElement root("Parameters");
	for (int i = 0; i < numParams; ++i)
	{
		root.setAttribute(Names[i], Params[i]);
	}
	copyXmlToBinary(root,destData);
}

void FnordAudioProcessor::setStateInformation(const void* data, int sizeInBytes)
{
	XmlElement * root = getXmlFromBinary(data,sizeInBytes);
	if (root == 0) {
		AlertWindow::showMessageBox(AlertWindow::InfoIcon,"failed to parse xml", "oh noes");
	} else {
		if (root->hasTagName("Parameters"))
		{
			for (int i = 0; i < numParams; ++i)
			{
				Params[i] = (float)root->getDoubleAttribute(Names[i]);
			}
		}
	}
}

//==============================================================================
// This creates new instances of the plugin..
AudioProcessor* JUCE_CALLTYPE createPluginFilter()
{
	return new FnordAudioProcessor();
}

//==============================================================================


void FnordAudioProcessor::processMIDI(bool& hasMsg, MidiMessage& midiMsg, int& index, int& i, MidiBuffer::Iterator& it, int poly)
{
	while (hasMsg && index == i)
	{
		if (midiMsg.isNoteOn())
		{
			++mNoteCount;

			bool found = false;
			int oldest = -1; uint64 lowest = mNoteCount;
			for (int v = 0; v < poly; ++v)
			{
				if (mOrder[v] < lowest)
				{
					lowest = mOrder[v];
					oldest = v;
				}
				if (mEnvStage[v] == Off)
				{
					mVelo[v] = midiMsg.getVelocity();
					mNote[v] = midiMsg.getNoteNumber();
					mEnvStage[v] = Att;
					mEnv[v] = 0;
					mOrder[v] = mNoteCount;
					found = true;
					break;
				}
			}
			// steal voice:  same note?
			if (!found)
			{
				for (int v = 0; v < poly; ++v)
				{
					if (mNote[v] == midiMsg.getNoteNumber())
					{
						mVelo[v] = midiMsg.getVelocity();
						mEnvStage[v] = Att;
						mOrder[v] = mNoteCount;
						found = true;
						break;
					}
				}
			}
			// steal voice: lowest in release stage?
			if (!found)
			{
				int tv = -1; float le = 1.0f;
				for (int v = 0; v < poly; ++v)
				{
					if (mEnvStage[v] == Rel)
					{
						float e = env(v);
						if (e < le)
						{
							le = e;
							tv = v;
							found = true;
						}
					}
				}
				if (found)
				{
					mVelo[tv] = midiMsg.getVelocity();
					mNote[tv] = midiMsg.getNoteNumber();
					mEnvStage[tv] = Att;
					mEnv[tv] = 0;
					mOrder[tv] = mNoteCount;
				}
			}
			// steal voice: oldest
			if (!found)
			{
				mVelo[oldest] = midiMsg.getVelocity();
				mNote[oldest] = midiMsg.getNoteNumber();
				mEnvStage[oldest] = Att;
				mEnv[oldest] = 0;
				mOrder[oldest] = mNoteCount;
			}
		}
		else if (midiMsg.isNoteOff())
		{
			for (int v = 0; v < VOICES; ++v)
			{
				if (mNote[v] == midiMsg.getNoteNumber() && mEnvStage[v] != Rel)
				{
					mEnvStage[v] = Rel;
				}
			}
		}
		hasMsg = it.getNextEvent(midiMsg, index);
	}
}

void FnordAudioProcessor::processBlock(AudioSampleBuffer& buffer, MidiBuffer& midiMessages)
{
	const float& vol = Params[Vol];
	uint8 poly = getPoly(); 
	const float& inter = Params[Interp];
	const float& change = (Params[Change]*Params[Change]);
	const uint8 tableSize = (uint8)(getSize(Params[Size])-1);

	MidiMessage midiMsg(0xf0);

	float* p0 = buffer.getWritePointer(0);
	float* p1 = buffer.getWritePointer(1);

	MidiBuffer::Iterator it(midiMessages);
	int index = 0;
	bool hasMsg = it.getNextEvent(midiMsg, index);

	for (int i = 0; i < buffer.getNumSamples(); ++i)
	{
		processMIDI(hasMsg, midiMsg, index, i, it, poly);

		float out=0;
		for (int v = 0; v < VOICES; ++v)
		{
			float envVal = env(v);
			if (envVal > 0)
			{
				float mHz = 220.0f * pow(2.0f, ((float)(mNote[v] - 9) / 12.0f));

				mPhase[v] += mHz;
				if (mPhase[v] >= 44100)
				{
					mStep[v]++;
					if (mStep[v] >= tableSize) {
						mStep[v] = 0;
					}
					mPhase[v] -= 44100;

					float next = 0;
					switch (getAlgo()) {
					case OneBit:
						next = (Random::getSystemRandom().nextFloat()>.95f) ? 1.0f : -1.0f;
						break;
					case Ramp:
						next = table[mStep[v]] + (change*.1f);
						if (next > 1.0f) { next -= 2.0f; }
						break;
					case Swoop:
						next = table[mStep[v]] + ((float)mStep[v]/(float)tableSize);
						while (next > 1.0f) { next -= 2.0f; }
						break;
					case Soft:
						{
							float a0 = table[mStep[v]];
							float a1 = getVal(mStep[v]+1, tableSize);
							if (fabsf(a1-a0) < .1f) 
							{
								next = (Random::getSystemRandom().nextFloat()*2.0f)-1.0f;
							} else {
								next = (a0 + a1) * .5f;
							}
						}
						break;
					case XOR:
						{
							next = getVal(mStep[v]+1, tableSize);
							uint8 t0 = (uint8)((table[mStep[v]] + 1) * 128.0f);
							uint8 t1 = (uint8)((next + 1) * 128.0f);
							uint8 tx = t0 ^ t1;
							next = ((float)tx / 128.0f) - 1.0f;
							break;
						}
					case Rot:
						{
							next = getVal(mStep[v]+1, tableSize);
							uint8 t0 = (uint8)((table[mStep[v]] + 1) * 128.0f);
							uint8 tx = (t0 << 1) | (t0 & 1);
							next = ((float)tx / 128.0f) - 1.0f;
							break;
						}
					case Tent:
						{
							float u = (1 + Params[Change]) * 1.99999f;
							next = (getVal(mStep[v], tableSize)+1) * .5f;
							if (next < .5f) {
								next = u * next;
							} else 
							{
								next = u * (1- next);
							}
							next = (next*2.0f) - 1.0f;
							while (next > 1.0f) next -= 2.0f;
							while (next < -1.0f) next += 2.0f;
							break;
						}
					case Unblur:
						{
							next = getVal(mStep[v]+1, tableSize);
							float prev = getVal(mStep[v]-1, tableSize);
							float avg = (next + prev) * .5f;
							next = (table[mStep[v]]*2) - avg;
							while (next > 1.0f) { next -= 2.0f; }
							while (next < -1.0f) { next += 2.0f; }
							break;
						}
					case Note:
						{
							uint8 t0 = (uint8)((table[mStep[v]] + 1) * 128.0f);
							uint8 t1 = (uint8)mOrder[v] << 1;
							uint8 tx = t0 ^ t1;
							next = ((float)tx / 128.0f) - 1.0f;
							break;
						}
					}					
					table[mStep[v]] = lerp(change, table[mStep[v]], next);  
				}

				float t = (mPhase[v] / 44100.0f);
				if (inter < .5f) {
					t *= (inter*2.0f);
				}
				else 
				{
					t = lerp((inter-.5f)*2, t, 1-t);
				}
				float myOut = lerp(t, getVal(mStep[v],tableSize), getVal(mStep[v]+1,tableSize));
				myOut *= (mVelo[v] / 127.0f);

				out += (myOut * envVal * envVal);
			}
		}
		out = out * vol;

		*p0 = out;
		*p1 = out;
		++p0;
		++p1;
	}
}

float FnordAudioProcessor::env(int voice) {
	double guard = 15;

	switch (mEnvStage[voice])
	{
		case Off:
		{
					mEnv[voice] = 0.0f;
					return 0.0f;
		}
		case Att:
		{
					double attackTime = .001f + (Params[Attack] * 44100.0) + guard;
					mEnv[voice] += (float)(1 / attackTime);
					if (mEnv[voice] >= 1.0f)
					{
						mEnv[voice] = 1.0f;
						mEnvStage[voice] = Sus;
					}
					return mEnv[voice];
					break;
		}
		case Sus:
		{
					float& sus = Params[Sustain];
					if (mEnv[voice] > sus)
					{
						double relTime = .0005f + (Params[Release] * 88200.0);
						mEnv[voice] -= (float)(1 / relTime);
						if (mEnv[voice] < sus)
						{
							mEnv[voice] = sus;
						}
					}
					return mEnv[voice];
		}
		case Rel:
		{
					double relTime = 20 + (Params[Release] * 88200.0) + guard;
					mEnv[voice] -= (float)(1 / relTime);
					if (mEnv[voice] < 0)
					{
						mEnv[voice] = 0;
						mEnvStage[voice] = Off;
					}
					return mEnv[voice];
		}
	}

	return 0.0f; // this shouldn't happen
}

float FnordAudioProcessor::getVal(int index, uint8 tableSize)
{
	int useIndex = index % tableSize;
	return table[useIndex];
}
