/*
  ==============================================================================

    This file was auto-generated!

    It contains the basic startup code for a Juce application.

  ==============================================================================
*/

#ifndef PLUGINPROCESSOR_H_INCLUDED
#define PLUGINPROCESSOR_H_INCLUDED

#include "../JuceLibraryCode/JuceHeader.h"


//==============================================================================
/**
*/
class FnordAudioProcessor  : public AudioProcessor
{
public:
	//==============================================================================
	FnordAudioProcessor();
	~FnordAudioProcessor();

	//==============================================================================
	void prepareToPlay(double sampleRate, int samplesPerBlock);
	void releaseResources();

	void processBlock(AudioSampleBuffer& buffer, MidiBuffer& midiMessages);

	//==============================================================================
	AudioProcessorEditor* createEditor();
	bool hasEditor() const;

	//==============================================================================
	const String getName() const;

	int getNumParameters();

	float getParameter(int index);
	void setParameter(int index, float newValue);

	const String getParameterName(int index);
	const String getParameterText(int index);

	const String getInputChannelName(int channelIndex) const;
	const String getOutputChannelName(int channelIndex) const;
	bool isInputChannelStereoPair(int index) const;
	bool isOutputChannelStereoPair(int index) const;

	bool acceptsMidi() const;
	bool producesMidi() const;
	bool silenceInProducesSilenceOut() const;
	double getTailLengthSeconds() const;

	//==============================================================================
	int getNumPrograms();
	int getCurrentProgram();
	void setCurrentProgram(int index);
	const String getProgramName(int index);
	void changeProgramName(int index, const String& newName);

	//==============================================================================
	void getStateInformation(MemoryBlock& destData);
	void setStateInformation(const void* data, int sizeInBytes);

	enum Parameters{ Interp, Size, Algo, Change, Attack, Sustain, Release, Poly, Vol, numParams };
	enum EnvStage{ Off = 0, Att, Sus, Rel };
	enum AlgoType { Soft=0, XOR, Unblur, OneBit, Ramp, Swoop, Rot, Tent, Note, numAlgos };

private:

	float env(int v);
	inline float limit(const float& f) 
	{
		if (f<0) return 0;
		if (f>1) return 1;
		return f;
	}
	inline float lerp(float t, float a, float b) {
		return (1 - t)*a + (t*b);
	}

#define MAX_TABLE 256
#define VOICES 4

	inline uint16 getSize(float f) {
		int i = 4 + (uint8)(f * 4.9999f);
		return 1 << i;
	}
	inline uint8 getPoly() {
		return (uint8)(1 + (Params[Poly]*3.999f));	
	}
	AlgoType getAlgo() {
		return (AlgoType)(int)(Params[Algo] * .999f * (float)numAlgos);
	}
	String getAlgoName() {
		AlgoType t = getAlgo();
		switch (t)
		{
			case Soft:
				return "Soft";
			case XOR:
				return "XOR";
			case Unblur:
				return "Unblur";
			case OneBit:
				return "1BitBias";
			case Ramp:
				return "Ramp";
			case Swoop:
				return "Swoop";
			case Rot:
				return "Rot";
			case Tent:
				return "Tent";
			case Note:
				return "Note";
		}
		return "???";
	}

	void processMIDI(bool& hasMsg, MidiMessage& msg, int& index, int& i, MidiBuffer::Iterator& it, int poly);
	float getVal(int, uint8);

	float Params[numParams];
	StringArray Names;
	double mSR;


	int mNote[VOICES];
	uint8 mVelo[VOICES];
	float mPhase[VOICES];
	uint8 mStep[VOICES];
	EnvStage mEnvStage[VOICES];
	float mEnv[VOICES];
	uint64 mOrder[VOICES];
	uint64 mNoteCount;
	float table[MAX_TABLE];


	//==============================================================================
	JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR(FnordAudioProcessor);
};

#endif  // PLUGINPROCESSOR_H_INCLUDED
